/* 
*** AVK 14/07/2014 notes:***
  
  2) Minimize temporary variables
  3) Change 3d FFT to 3 1d FFT
  4) Fix diagnostics
  5) Provide the correct example input file for trunk and phase mixing branch
  6) Check how beta appears in normalization
  7) Write up JCP

  timestep.cu:
  1) Try and see if force_d can be removed. 
***AVK 25/07/2013 notes:***

 1) Write destroy_arrays functions

 3) Start fixing diagnostics/flow:
 	a) Define diagnostic calls depending on solve_alfven, solve_slow etc.
	b) Write a separate allocate/free file
	c) Rewrite Alfven diagnostics: think of minimizing arrays
	d) Rewrite slow diagnostics: think of minimizing arrays

 4) Do time advances in this order:
 	a) time_adv_phase_mix
	b) time_adv_alf : Done
	c) time_adv_full

 5) Build a test suite : Physics tests

***************
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <cufft.h>
#include <ctime>
#include <netcdf.h> //GGH: NetCDF Library

/////////////////////////////
// Input parameters
/////////////////////////////
//Device
int devid;

// algorithm choices
bool debug, restart;
int nwrite, nforce; // Set nforce very large if driven is false
float maxdt, cfl;

// Initial conditions
bool decaying, driven, orszag_tang, noise;
float aw_coll;

//Computation Grid
__constant__ int Nx, Ny, Nz, zThreads;
__constant__ int Nm;
__constant__ float X0, Y0, Z0;
int nsteps;
float endtime;

// Grid, block setup for GPU
dim3 dimGrid, dimBlock;
int totalThreads;

// forcing
int nkstir;
int *kstir_x, *kstir_y, *kstir_z;
int gm_nkstir;
int *gm_kstir_x, *gm_kstir_y, *gm_kstir_z;
float fampl, gm_fampl;

// dissipation
int alpha_z, alpha_hyper;
float nu_kz, nu_hyper;

// Internal variables

char *runname;
char *restartname;
char stopfile[255];
bool file_exists(char *filename);
void read_namelist(char *filename);
void restartRead(cuComplex* zp, cuComplex* zm, float* tim);
void restartWrite(cuComplex* zp, cuComplex* zm, float tim);
FILE *energyfile, *alf_kzkpfile, *awp_collfile, *awm_collfile;
int ncid=-99;  // ID for NetCDF frequency diagnostic file
int time_freq_id, zp_freq_id, zm_freq_id;

//Dummy arrays
cuComplex *temp1, *temp2, *temp3, *temp4;
cuComplex *padded;
cuComplex *dx, *dy;
float *fdxR, *fdyR, *gdxR, *gdyR;

cufftHandle plan_C2R, plan_R2C, plan2d_C2R;


//////////////////////////////////////////////////////////////////////
// Include files
//////////////////////////////////////////////////////////////////////
#include "c_fortran_namelist3.c"
#include "device_funcs.cu"
#include "k_funcs.cu"
#include "zderiv_kernel.cu"
#include "reduc_kernel.cu"
#include "work_kernel.cu"
#include "maxReduc.cu"
#include "sumReduc_nopad.cu"
#include "init_kernel.cu"
#include "init_func.cu"
#include "nlps_kernel.cu"
#include "timestep_kernel.cu"
#include "diag_kernel.cu"
#include "damping_kernel.cu"
#include "nlps.cu"
#include "freq_diag.cu"
#include "diagnostics.cu"
#include "nonlin.cu"
#include "courant.cu"
#include "forcing.cu"
#include "timestep.cu"

// Declare all required arrays
// Declare host variables
cuComplex *f, *g;

// Declare device variables
cuComplex *f_d, *g_d;
cuComplex *f_d_tmp, *g_d_tmp;


// Declare functions
void allocate_arrays();
void destroy_arrays();
void setup_device();
void setup_grid_block();

//////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
  if ( argc < 1 ) 
    {
      printf( "Usage: ./gandalf runname");
    }
  else 
    {
	  // Assuming argv[1] is the runname
	  runname = argv[1];
	  char str[255];
	  strcpy(str, runname);
	  strcat(str, ".in");
	  printf("Reading from %s \n", str);
      // Read namelist
	  read_namelist(str);

      cudaEvent_t start, stop;
      cudaEventCreate(&start);
      cudaEventCreate(&stop);

      cudaEventRecord(start, 0);
  

      // Allocate arrays
      allocate_arrays();

	  // Setup dimgrid and dimblock
      setup_grid_block();

	  // Stopfile
      strcpy(stopfile, runname);
      strcat(stopfile, ".stop");

	  // Initialize diagnostics
	  init_diag();

	  // Create FFT plans
	  fft_plan_create();

	//////////////////////////////////////////////////////////
    float dt = 1.e-2;
    float tim=0;
    int istep=0;
    srand( (unsigned) time(NULL));
	//////////////////////////////////////////////////////////
	//If not restarting, initialize
	if(!restart){
		if(debug) printf("Not restarting. \n");

		//////////////////////////////////////////////////////////
		// Initialize Phi and Psi
		finit(f,g);
		
        // Transfer the fields to Device
        cudaMemcpy(f_d, f, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz, cudaMemcpyHostToDevice);
        if(debug) printf("f Initialization on device: %s\n",cudaGetErrorString(cudaGetLastError()));
        cudaMemcpy(g_d, g, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz, cudaMemcpyHostToDevice);
        if(debug) printf("g Initialization on device: %s\n",cudaGetErrorString(cudaGetLastError()));
     }
	// If restarting:
	else{
		if(debug) printf("Restarting. \n");
		restartRead(f_d, g_d, &tim);

		printf("Time after restart: %f \n", tim);
	}

	//////////////////////////////////////////////////////////

	//////////////////////////////////////////////////////////

    // Zeroth step
	//    diagnostics(f_d, g_d, tim, 0);
	//    courant(&dt, f_d, g_d);
	//    printf("dt = %f\n", dt);
	//    advance(f_d_tmp, f_d, g_d_tmp, g_d, dt, istep);
    // GGH NOTE: I added tim+=dt and swapped order of istep++ to be consistent with main loop
	//    tim+=dt;
	//    istep++;
    //    diagnostics(f_d, g_d, tim, 0);

	//GGH NOTE: I reordered these statements so that diagnostics are output before
        // any steps have been taken, and then the fields are advanced before diagnostics.
	
	
    // Output diagnostics for initial condition
    courant(&dt, f_d, g_d);
    printf("dt = %f\n", dt);
    diagnostics(f_d, g_d, tim, 0);

    while(istep < nsteps) {

		// Check CFL condition
		courant(&dt, f_d, g_d);

		// Time advance RMHD & slow mode equations
		advance(f_d_tmp, f_d, g_d_tmp, g_d, dt, istep);

 		tim+=dt;
		istep++;
           
		if(istep % nwrite ==0) {
			printf("%f      %d\n",tim,istep);
			printf("dt = %f\n",dt);
			diagnostics(f_d, g_d, tim,istep/nwrite);
			printf("%s\n",cudaGetErrorString(cudaGetLastError()));
			}

		// Check if stopfile exists.
		if(file_exists(stopfile)) {
			printf("Simulation stopped by user with stopfile \n");
			break;
		}
	
  } 

	//diagnostics(tim);

	if(debug) printf("Before restart write \n");
	restartWrite(f_d, g_d, tim);
    if(debug) printf("Restart write done \n");
    printf("Done.\n");
	
	
	float elapsed_time;
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&elapsed_time, start, stop);
	printf("Total time(ms) = %f \n", elapsed_time);
	printf("Average time (ms) = %f \n", elapsed_time/istep);

	////////////////////
	// Clean up Area
	////////////////////

	// Destroy fft plan and close diagnostics
	fft_plan_destroy(); 
	close_diag();

	// Free all arrays
	destroy_arrays();
	if(driven) {free(kstir_x); free(kstir_y); free(kstir_z);}

	}  	
	return 0;
}


bool file_exists(char *filename){
	if(FILE *file = fopen(filename, "r"))
	{
		fclose(file);
		return true;
	 }
	return false;
}

void read_namelist(char* filename)
{
  struct fnr_struct namelist_struct = fnr_read_namelist_file(filename);

  int debug_i, restart_i;
  int decaying_i, driven_i, orszag_tang_i, noise_i;
  
  //Device
  if(fnr_get_int(&namelist_struct, "dev", "devid", &devid)) devid = 0;
  setup_device();
  // algo 
  if(fnr_get_int(&namelist_struct, "algo", "debug", &debug_i)) debug_i=0;
   if(debug_i == 0) { debug = false; }
   else {debug = true;}
  if(fnr_get_int(&namelist_struct, "algo", "restart", &restart_i)) restart_i = 0;
   if(restart_i == 0) { restart = false;}
   else {restart = true;}
  if(restart) fnr_get_string(&namelist_struct, "algo", "rest", &restartname);
  if(fnr_get_int(&namelist_struct, "algo", "nwrite", &nwrite)) nwrite = 1;
  if(fnr_get_int(&namelist_struct, "algo", "nforce", &nforce)) nforce = 1;
  if(fnr_get_float(&namelist_struct, "algo", "maxdt", &maxdt)) maxdt = .1;
  if(fnr_get_float(&namelist_struct, "algo", "cfl", &cfl)) cfl = .1;

  if(debug) printf("Algo read \n");

  // Initial conditions
  fnr_get_int(&namelist_struct, "init", "decaying", &decaying_i);
   if(decaying_i == 0) { decaying = false; }
   else {decaying = true;}
  fnr_get_int(&namelist_struct, "init", "driven", &driven_i);
   if(driven_i == 0) { driven = false; }
   else {driven = true;}
   fnr_get_int(&namelist_struct, "init", "orszag_tang", &orszag_tang_i);
   if(orszag_tang_i == 0) { orszag_tang = false; }
   else {orszag_tang = true;}
   if(orszag_tang) printf("orszag_tang \n");
   fnr_get_int(&namelist_struct, "init", "noise", &noise_i);
   if(noise_i == 0) { noise = false; }
   else {noise = true;}
   if(fnr_get_float(&namelist_struct, "init", "aw_coll", &aw_coll)) aw_coll = 0 ;

  if(debug) printf("Initial conditions read \n");

  //Grid
  if(fnr_get_int(&namelist_struct, "grid", "Nx", &Nx)) *&Nx = 16;
  cudaMemcpyToSymbol(Nx, &Nx, sizeof(int));
  if(fnr_get_int(&namelist_struct, "grid", "Ny", &Ny)) *&Ny = 16;
  cudaMemcpyToSymbol(Ny, &Ny, sizeof(int));
  if(fnr_get_int(&namelist_struct, "grid", "Nz", &Nz)) *&Nz = 16;
  cudaMemcpyToSymbol(Nz, &Nz, sizeof(int));
  if(fnr_get_float(&namelist_struct, "grid", "X0", &X0)) *&X0 = 1.0f;
  cudaMemcpyToSymbol(X0, &X0, sizeof(float));
  if(fnr_get_float(&namelist_struct, "grid", "Y0", &Y0)) *&Y0 = 1.0f;
  cudaMemcpyToSymbol(Y0, &Y0, sizeof(float));
  if(fnr_get_float(&namelist_struct, "grid", "Z0", &Z0)) *&Z0 = 1.0f;
  cudaMemcpyToSymbol(Z0, &Z0, sizeof(float));
  if(fnr_get_int(&namelist_struct, "grid", "nsteps", &nsteps)) {
    printf("nsteps not specified. Exiting ... \n");
    exit(1);
    }

  if(debug) printf("Grid read \n");

  // Dissipation
  if(fnr_get_int(&namelist_struct, "dissipation", "alpha_z", &alpha_z)) alpha_z = 2;
  if(fnr_get_float(&namelist_struct, "dissipation", "nu_kz", &nu_kz)) nu_kz = 1.0f;
  if(fnr_get_int(&namelist_struct, "dissipation", "alpha_hyper", &alpha_hyper)) alpha_hyper = 2;
  if(fnr_get_float(&namelist_struct, "dissipation", "nu_hyper", &nu_hyper)) nu_hyper = 1.0f;

  if(debug) printf("Dissipation read \n");

  // Forcing
   fnr_get_int(&namelist_struct, "forcing", "nkstir", &nkstir);
   fnr_get_float(&namelist_struct, "forcing", "fampl", &fampl);

   //   int tmpint=0;
   //  fnr_get_int(&namelist_struct, "forcing", "nkstir", &tmpint);
   //  printf("tmpint= %i \n",tmpint);

   if(driven){

		kstir_x = (int*) malloc(sizeof(int)*nkstir);
		kstir_y = (int*) malloc(sizeof(int)*nkstir);
		kstir_z = (int*) malloc(sizeof(int)*nkstir);


 // Initialize Forcing modes
   char tmp_str[255];
   char buffer[20];
   int f_k;
   for(int ikstir=0; ikstir<nkstir; ikstir++) {
     
     strcpy(tmp_str,"stir_");
     sprintf(buffer,"%d", ikstir);
	 strcat(tmp_str, buffer);
	 printf("%s  \n",tmp_str);  // GGH
	 fnr_get_int(&namelist_struct, tmp_str, "kx", &f_k);
	 printf("fk= %i \n",f_k);
	 kstir_x[ikstir] = (f_k + Nx) % Nx;
	 fnr_get_int(&namelist_struct, tmp_str, "ky", &f_k);
	 kstir_y[ikstir] = (f_k + Ny) % Ny;
	 fnr_get_int(&namelist_struct, tmp_str, "kz", &f_k);
	 kstir_z[ikstir] = (f_k + Nz) % Nz;

	 printf("kx= %i,   ky= %i,   kz= %i  \n",kstir_x[ikstir],kstir_y[ikstir],kstir_z[ikstir] );  


    }
   }

  if(debug) printf("Forcing read \n");

}


//////////////////////////////////////////////////////////////////////
// Restart routines
//////////////////////////////////////////////////////////////////////

void restartWrite(cuComplex* zp, cuComplex* zm, float tim)
{
  if(debug) printf("Entering restart write\n");
  char str[255];
  FILE *restart;
  strcpy(str, runname);
  strcat(str,".res");
  restart = fopen(str, "wb");
  if(debug) printf("Opened restart file to write\n");

  cuComplex *zp_h;
  cuComplex *zm_h;
  if(debug) printf("Declared arrays \n");
  
  if(debug) printf("Nx = %d, Ny = %d, Nz = %d, Nm = %d \n", Nx, Ny, Nz, Nm);

  zp_h = (cuComplex*) malloc(sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
  cudaMemcpy(zp_h, zp, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz, cudaMemcpyDeviceToHost);
  if(debug) printf("Copying over zp %s\n",cudaGetErrorString(cudaGetLastError()));
  zm_h = (cuComplex*) malloc(sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
  cudaMemcpy(zm_h, zm, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz, cudaMemcpyDeviceToHost);
  if(debug) printf("Copying over zm %s\n",cudaGetErrorString(cudaGetLastError()));
  if(debug) printf("Allocated arrays \n");
  
  fwrite(&tim,sizeof(float),1,restart);
  if(debug) printf("Wrote istep, tim \n");
  
  fwrite(zp_h,sizeof(cuComplex)*Nx*(Ny/2+1)*Nz,1,restart);
  if(debug) printf("Wrote zp \n");
  fwrite(zm_h,sizeof(cuComplex)*Nx*(Ny/2+1)*Nz,1,restart);
  if(debug) printf("Wrote zm \n");
  
  
  fclose(restart);

  free(zp_h); free(zm_h); 
  
  
}


void restartRead(cuComplex* zp, cuComplex* zm, float* tim)
{
  char str[255];
  FILE *restart;
  strcpy(str, restartname);
  strcat(str,".res");
  if(debug) printf("Restartfile = %s \n", str);
  restart = fopen(str, "rb");

  cuComplex *zp_h;
  cuComplex *zm_h;
  
  zp_h = (cuComplex*) malloc(sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
  zm_h = (cuComplex*) malloc(sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);

  fread(tim,sizeof(float),1,restart);

  fread(zp_h, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz, 1, restart);
  cudaMemcpy(zp, zp_h, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz, cudaMemcpyHostToDevice);
  fread(zm_h, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz, 1, restart);
  cudaMemcpy(zm, zm_h, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz, cudaMemcpyHostToDevice);

  fclose(restart);

  free(zp_h);
  free(zm_h);
  
  
}


// Setup device
void setup_device(){

  // Device information
  int ct, dev;
  struct cudaDeviceProp prop;

  cudaGetDeviceCount(&ct);
  printf("Device Count: %d\n",ct);

  cudaSetDevice(devid);
  cudaGetDevice(&dev);
  printf("Device ID: %d\n",dev);

  cudaGetDeviceProperties(&prop,dev);
  printf("Device Name: %s\n", prop.name);

  printf("Major mode: %d\n", prop.major);
  printf("Global Memory (bytes): %lu\n", (unsigned long) prop.totalGlobalMem);
  printf("Shared Memory per Block (bytes): %lu\n", (unsigned long) prop.sharedMemPerBlock);
  printf("Registers per Block: %d\n", prop.regsPerBlock);
  printf("Warp Size (threads): %d\n", prop.warpSize); 
  printf("Max Threads per Block: %d\n", prop.maxThreadsPerBlock);
  printf("Max Size of Block Dimension (threads): %d * %d * %d\n", prop.maxThreadsDim[0], 
	 prop.maxThreadsDim[1], prop.maxThreadsDim[2]);
  printf("Max Size of Grid Dimension (blocks): %d * %d * %d\n", prop.maxGridSize[0], prop.maxGridSize[1], prop.maxGridSize[2]);

  }
void setup_grid_block(){
  int dev;
  struct cudaDeviceProp prop;
  cudaGetDevice(&dev);
  cudaGetDeviceProperties(&prop,dev);

  //////////////////////////////////////////////////////////
	//set up normal dimGrid/dimBlock config  
	int zBlockThreads = prop.maxThreadsDim[2];
	*&zThreads = zBlockThreads*prop.maxGridSize[2];
	totalThreads = prop.maxThreadsPerBlock;

	if(Nz > zBlockThreads) dimBlock.z = zBlockThreads;
	else dimBlock.z = Nz;
	int xy = (int) totalThreads/dimBlock.z;
	int blockxy = (int) sqrt((float) xy);
	//dimBlock = threadsPerBlock, dimGrid = numBlocks
	dimBlock.x = blockxy;
	dimBlock.y = blockxy;

	if(Nz>zThreads) {
	  dimBlock.x = (unsigned int) sqrt((float) totalThreads/zBlockThreads);
	  dimBlock.y = (unsigned int) sqrt((float) totalThreads/zBlockThreads);
	  dimBlock.z = zBlockThreads;
	}  

	dimGrid.x = (unsigned int) ceil((float) Nx/dimBlock.x + 0);
	dimGrid.y = (unsigned int) ceil((float) Ny/dimBlock.y + 0);
	if(prop.maxGridSize[2]==1) dimGrid.z = 1;    
	else dimGrid.z = (unsigned int) ceil((float) Nz/dimBlock.z) ;
	cudaMemcpyToSymbol(zThreads, &zThreads, sizeof(int));
	printf("zthreads = %d, zblockthreads = %d \n", zThreads, zBlockThreads);
	
	printf("dimGrid = %d, %d, %d \t dimBlock = %d, %d, %d \n", dimGrid.x, dimGrid.y, dimGrid.z, dimBlock.x, dimBlock.y, dimBlock.z);

	}

////////////////////////////////////////
// Array allocation/destruction functions
////////////////////////////////////////
// Allocate arrays
void allocate_arrays()
 {

      printf("Allocating arrays...\n");
      // Allocate host arrays
      f = (cuComplex*) malloc(sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
      g = (cuComplex*) malloc(sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);

      // Allocate device arrays
      cudaMalloc((void**) &f_d, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
      cudaMalloc((void**) &g_d, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
      cudaMalloc((void**) &f_d_tmp, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
      cudaMalloc((void**) &g_d_tmp, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);

      cudaMalloc((void**) &temp1, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
      cudaMalloc((void**) &temp2, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
      cudaMalloc((void**) &temp3, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
      cudaMalloc((void**) &temp4, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);

      cudaMalloc((void**) &fdxR, sizeof(float)*Nx*Ny*Nz);
      cudaMalloc((void**) &fdyR, sizeof(float)*Nx*Ny*Nz);
      cudaMalloc((void**) &gdxR, sizeof(float)*Nx*Ny*Nz);
      cudaMalloc((void**) &gdyR, sizeof(float)*Nx*Ny*Nz);

	  cudaMalloc((void**) &dx, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);
	  cudaMalloc((void**) &dy, sizeof(cuComplex)*Nx*(Ny/2+1)*Nz);

      cudaMalloc((void**) &padded, sizeof(cuComplex)*Nx*Ny*Nz);

 }
void destroy_arrays(){

	// Destroy host arrays
	free(f); free(g);

	// Destroy device arrays containing fields
	cudaFree(f_d); cudaFree(g_d);


	// Destroy dummy arrays
		// Fields
	cudaFree(f_d_tmp); cudaFree(g_d_tmp);

		// nonlin
	cudaFree(temp1); cudaFree(temp2); cudaFree(temp3); cudaFree(temp4);

	cudaFree(fdxR); cudaFree(fdyR); cudaFree(gdxR); cudaFree(gdyR);

	cudaFree(dx); cudaFree(dy);

		// courant
	cudaFree(padded);
}


